# usepandoc.py ver 0.8 (encoding: utf-8)  2020/11/28
  # Copyright (C) T. Yoshiizumi, 2019-2020 All rights reserved.
import os, sys, platform
import subprocess
import re
import glob
import copy
from collections import OrderedDict
from chardet.universaldetector import UniversalDetector
from platform import python_version
import docx
from docx.shared import Inches
from docx.shared import Mm
from docx.shared import Pt
from docx.enum.style import WD_STYLE_TYPE
from docx.enum.text import WD_TAB_ALIGNMENT
from docx.enum.section import WD_ORIENT
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.enum.table import WD_TABLE_ALIGNMENT
from docx.oxml import OxmlElement
from docx.oxml.ns import qn
from docx.text.run import Run
from docx.oxml.text.run import CT_R

# global variables
_default_charset = 'utf-8'
if platform.system() == 'Windows':
    _default_charset = 'cp932'
_charset = _default_charset
if int(python_version()[0]) < 3:
    reload(sys)
    sys.setdefaultencoding(_charset)
_basedir = os.path.dirname(os.path.abspath(__file__))
    # options for Word
_optKeys = ["TableBorder", "TableCenter", "TableRow1",
    "FontSize", "PageSize", "PageNumber",
    "TextColumns", "CharsLine", "LinesPage",
    "ImageCenter", "portrait", "RemakeAfter"]
_optHash = {"TableBorder": True, "TableCenter": True, "TableRow1": True,
    "FontSize": 12.0, "PageSize": "A4", "PageNumber": True,
    "TextColumns": 1, "CharsLine": 0, "LinesPage": 0,
    "ImageCenter": True, "portrait": True, "RemakeAfter": True}
    # page size
_sizeDict = {'B4': [257, 364],
    'A4': [210, 297],
    'B5': [182, 257],
    'A5': [148, 210],
    'HG': [100, 148]}
u_dict = {'0':'none', '1':'single', '2':'words', '3':'double',
    '4':'dotted', '6':'thick', '7':'dash', '9':'dotDash', '10':'dotDotDash',
    '11':'wave', '20':'dottedHeavy', '23':'dashedHeavy',
    '25':'dashDotHeavy', '26':'dashDotDotHeavy', '27':'wavyHeavy',
    '39':'dashLong', '43':'wavyDouble', '55':'dashLongHeavy'}
em_dict = {'0':'none', '1':'dot', '2':'comma', '3':'circle', '4':'underDot'}

# sets global values (_charset, _default_charset)
def set_charset(charset = None, default_charset = None):
    if charset is not None:
        global _charset
        _charset = charset
    if default_charset is not None:
        global _default_charset
        _default_charset = default_charset
    if int(python_version()[0]) < 3:
        reload(sys)
        sys.setdefaultencoding(_charset)

# gets the file's charset(encoding)
def get_charset(fname):
    detector = UniversalDetector()
    f = open(fname, mode='rb')
    for binary in f:
        detector.feed(binary)
        if detector.done:
            break
    f.close()
    detector.close()
    enc = detector.result['encoding']
    if enc is None:
        enc = _default_charset
    return enc

# read text file (return unicode string)
def read_text(filename, charset=''):
    try:
        flag = os.path.exists(filename)
    except:
        flag = False
    if not flag:
        return None
    if charset == '':
        charset = get_charset(filename)
    if int(python_version()[0]) < 3:
        f = open(filename, mode='r')
        data = f.read()
        f.close()
        data = unicode(data, charset)
    else:
        f = open(filename, mode='rt', encoding=charset)
        data = f.read()
        f.close()
    return data

# write a text to a file
def write_text(filename, text, charset=''):
    if charset == '':
        charset = _charset
    if int(python_version()[0]) < 3:
        text = text.encode(charset)
        with open(filename, "w") as f:
            f.write(text)
    else:
        with open(filename, "w", encoding=charset) as f:
            f.write(text)

# get options in a markdown file
def get_options(filename):
    str = read_text(filename)
    ptn = re.compile(r"^<!-- +(.+?) +-->", flags=(re.MULTILINE | re.DOTALL))
    mlist = re.findall(ptn, str)
    for i in range(len(mlist)):
        w = mlist[i]
        w = re.sub(r'\s+', " ", w)
        mlist[i] = w
    ptnPandoc = re.compile(r"^pandoc +")
    ptns = [re.compile(r'(PageSize)=(B4L?|A4L?|B5L?|A5L?|HGL?)'),
        re.compile(r'(FontSize)=([0-9.]+)'),
        re.compile(r'(CharsLine|LinesPage|TextColumns)=(\d+)'),
        re.compile(r'(TableBorder|TableCenter|TableRow1|' +
            r'PageNumber|ImageCenter|RemakeAfter)=(yes|no)')]
    optdct = copy.copy(_optHash)
    optdct['pandoc'] = ''
    for str in mlist:
        m = re.search(ptnPandoc, str)
        if m:
            w = re.sub(ptnPandoc, "", str)
            if optdct['pandoc'] != '':
                optdct['pandoc'] = optdct['pandoc'] + " "
            optdct['pandoc'] = optdct['pandoc'] + w
            continue
        for ptn in ptns:
            while True:
                m = re.search(ptn, str)
                if not m:
                    break
                key = m.group(1)
                val = m.group(2)
                if val in "yes no":
                    if val == 'yes':
                        val = True
                    else:
                        val = False
                elif key == 'FontSize':
                    val = float(val)
                elif key in "CharsLine LinesPage TextColumns":
                    val = int(val)
                elif key == 'PageSize':
                    if val[-1] == 'L':
                        optdct['portrait'] = False
                        val = val[0:-1]
                    else:
                        optdct['portrait'] = True
                optdct[key] = val
                str = str[0:m.start()] + str[m.end():len(str)]
    return optdct

# add heading ID(s) to the markdown string
def add_heading_id(str):
    lines = str.split("\n")
    ptnHeading1 = re.compile(r'^(#+ +.+)\{([^}]+)\} *$')
    ptnHeading2 = re.compile(r'^#+ +')
    ptnBlock = re.compile(r'^(~~~+|```+)')
    inBlock = 0
    count = 0
    for i in range(len(lines)):
        line = lines[i]
        m = re.search(ptnBlock, line)
        if m:
            if inBlock == 0:
                fence = m.group(1)
                inBlock = 1
                continue
            if line == fence:
                inBlock = 0
                continue
            continue
        if inBlock == 1:
            continue
        m = re.search(ptnHeading1, line)
        if m:
            g1 = m.group(1)
            g2 = m.group(2)
            if g2 == '-':
                g2 = ".unnumbered"
            if not "#" in g2:
                count = count + 1
                line = g1 + " {#session-ex%d %s}" % (count, g2)
                lines[i] = line
                continue
            continue
        m = re.search(ptnHeading2, line)
        if m:
            count = count + 1
            line = line + " {#session-ex%d}" % count
            lines[i] = line
    if count > 0:
        str = "\n".join(lines)
    return str

# set the border of a cell in a table
def set_cell_border(cell, **kwargs):
    """
    Set cell`s border
    Usage:
    set_cell_border(
        cell,
        top={"sz": 12, "val": "single", "color": "#FF0000", "space": "0"},
        bottom={"sz": 12, "color": "#00FF00", "val": "single"},
        start={"sz": 24, "val": "dashed", "shadow": "true"},
        end={"sz": 12, "val": "dashed"},
    )
    """
    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()
        # check for tag existnace, if none found, then create one
    tcBorders = tcPr.first_child_found_in("w:tcBorders")
    if tcBorders is None:
        tcBorders = OxmlElement('w:tcBorders')
        tcPr.append(tcBorders)
        # list over all available tags
    for edge in ('start', 'top', 'end', 'bottom', 'insideH', 'insideV'):
        edge_data = kwargs.get(edge)
        if edge_data:
            tag = 'w:{}'.format(edge)
                # check for tag existnace, if none found, then create one
            element = tcBorders.find(qn(tag))
            if element is None:
                element = OxmlElement(tag)
                tcBorders.append(element)
                # looks like order of attributes is important
            for key in ["sz", "val", "color", "space", "shadow"]:
                if key in edge_data:
                    element.set(qn('w:{}'.format(key)), str(edge_data[key]))

# arrange a table (TableCenter, TableRow1, TableBorder)
def arrange_table(tbl, optdct):
    std_bdr = {
        'top': {"sz": 8, "val": "single", "color": "#000000"},
        'bottom': {"sz": 8, "val": "single", "color": "#000000"},
        'start': {"sz": 8, "val": "single", "color": "#000000"},
        'end': {"sz": 8, "val": "single", "color": "#000000"}}
    row_c = len(tbl.rows)
    col_org = len(tbl.columns)
    col_c = col_org
    if col_c == 0:
        i = 0
        total_c = 1
        while True:
            try:
                w = tbl.cell(0, i).text
            except:
                total_c = i
                break
            i = i + 1
        col_c = int(total_c / row_c)
    else:
        total_c = row_c * col_c
    if optdct['TableCenter'] == True:
        tbl.alignment = WD_TABLE_ALIGNMENT.CENTER
    if optdct['TableRow1'] == True:
        for k in range(col_c):
            tbl.cell(0, k).paragraphs[0].alignment = \
                WD_ALIGN_PARAGRAPH.DISTRIBUTE
    if optdct['TableBorder'] == True:
        for k in range(total_c):
            i = int(k / col_c)
            j = k % col_c
            if col_org == 0:
                cell = tbl.cell(0, k)
            else:
                cell = tbl.cell(i, j)
            bdr = copy.deepcopy(std_bdr)
            if i == 0:
                bdr['top']['val'] = "double"  # single, double, dotDotDash
            elif i == (row_c-1):
                bdr['bottom']['val'] = "double"
            if j == 0:
                bdr['start']['val'] = "double"
            elif j == (col_c-1):
                bdr['end']['val'] = "double"
            set_cell_border(cell, **bdr)

# function: insert a new run just after the run
def insert_next(p, run):
    new_run_element = p._element._new_r()
    run._element.addnext(new_run_element)
    return Run(new_run_element, run._parent)

# function: divide a run ('part0 #(part1){ part2 } part3')
def ueRun_divide(p, run, ptn):
    rtext = run.text
    rstyle = run.style
    m = re.search(ptn, rtext)
    if m is None:
        return
    if re.search(r'^[ue]\d*$', m.groups()[1]) is None:
        return
    if m.groups()[0] == '' and m.groups()[3] == '':
        run.text = m.groups()[2]
        run.style = rstyle
        set_prop_ue(m.groups()[1], run)
        return
    if m.groups()[0] == '':
        new_run = insert_next(p, run)
        new_run.text = m.groups()[3]
        new_run.style = rstyle
        run.text = m.groups()[2]
        run.style = rstyle
        set_prop_ue(m.groups()[1], run)
        ueRun_divide(p, new_run, ptn)
        return
    new_run = insert_next(p, run)
    new_run.text = m.groups()[2]
    new_run.style = rstyle
    run.text = m.groups()[0]
    run.style = rstyle
    set_prop_ue(m.groups()[1], new_run)
    if m.groups()[3] != '':
        run = new_run
        new_run = insert_next(p, run)
        new_run.text = m.groups()[3]
        new_run.style = rstyle
        ueRun_divide(p, new_run, ptn)

# function: divide a run ('part0 #(part1){ part2')
def ueRun_divide1(p, rn, ptn):
    run = p.runs[rn]
    rtext = run.text
    rstyle = run.style
    m = re.search(ptn, rtext)
    if m.groups()[0] == '' and m.groups()[2] == '':
        p._p.remove(run._r)
        return m.groups()[1], "delete"
    if m.groups()[0] == '':
        run.text = m.groups()[2]
        run.style = rstyle
        return m.groups()[1], "after"
    if m.groups()[2] == '':
        run.text = m.groups()[0]
        run.style = rstyle
        return m.groups()[1], "before"
    new_run = insert_next(p, run)
    new_run.text = m.groups()[2]
    new_run.style = rstyle
    run.text = m.groups()[0]
    run.style = rstyle
    return m.groups()[1], "both"

# function: divide a run ('part0 } part1')
def run_divide2(p, rn, ptn):
    run = p.runs[rn]
    rtext = run.text
    rstyle = run.style
    m = re.search(ptn, rtext)
    if m.groups()[0] == '' and m.groups()[1] == '':
        p._p.remove(run._r)
        return "delete"
    if m.groups()[0] == '':
        run.text = m.groups()[1]
        run.style = rstyle
        return "after"
    if m.groups()[1] == '':
        run.text = m.groups()[0]
        run.style = rstyle
        return "before"
    new_run = insert_next(p, run)
    new_run.text = m.groups()[1]
    new_run.style = rstyle
    run.text = m.groups()[0]
    run.style = rstyle
    return "both"

# function: set properties (UnderLine or EmphasisMark)
def set_prop_ue(mark, run):
    m = re.search(r'([ue])(\d*)', mark)
    if m is None:
        return
    k = m.groups()[0]
    v = m.groups()[1]
    if k == 'u':  # UnderLine
        if v == '' or (v in u_dict) == False:
            v = '1'
        v = u_dict[v]
    else:  # EmphasisMark
        k = 'em'
        if v == '' or (v in em_dict) == False:
            v = '1'
        v = em_dict[v]
    obj = None
    for rPr in run._r.xpath('./w:rPr'):
        objs = rPr.xpath('./w:' + k)
        if len(objs) > 0:
            obj = objs[0]
            break
    if obj is None:
        ue = docx.oxml.shared.OxmlElement('w:' + k)
        ue.set(docx.oxml.shared.qn('w:val'), v)
        rPr.append(ue)
    else:
        obj.set(qn('w:val'), v)

# function: set UnderLine or EmphasisMark on the whole document
def attach_ue(document):
    ptn = re.compile(r'(.*?)#([ue]\d*)\{([^{}]+?)\}(.*)')
    ptn1 = re.compile(r'(.*?)#([ue]\d*)\{([^{}]*)')
    ptn2 = re.compile(r'([^{}]*?)\}(.*)')
    for p in document.paragraphs:
        if re.search(ptn, p.text) is None:
            continue
        for run in p.runs:
            ueRun_divide(p, run, ptn)
        while re.search(ptn, p.text) is not None:
            n1 = -1
            for i in range(len(p.runs)):
                if re.search(ptn1, p.runs[i].text) is not None:
                    n1 = i
                    break
            if n1 == -1:
                break
            for i in range((n1+1), len(p.runs)):
                if re.search(ptn2, p.runs[i].text):
                    n2 = i
                    break
            dtype2 = run_divide2(p, n2, ptn2)
            mark, dtype1 = ueRun_divide1(p, n1, ptn1)
            if dtype2 == "delete" or dtype2 == "after":
                n2 = n2 - 1
            if dtype1 == "delete":
                n2 = n2 - 1
            elif dtype1 == "before":
                n1 = n1 + 1
            elif dtype1 == "both":
                n1 = n1 + 1
                n2 = n2 + 1
            for i in range(n1, n2+1):
                set_prop_ue(mark, p.runs[i])

# function: divide a run ('part0 #[pr]{ part1 } part2')
def rubyRun_divide(p, run, ptn):
    rtext = run.text
    rstyle = run.style
    m = re.search(ptn, rtext)
    if m is None:
        return
    baseText = m.groups()[1]
    rubyText = re.sub(r'[ -/:-@{-~]+', '', m.groups()[2])
    if m.groups()[0] == '' and m.groups()[3] == '':
        run.text = baseText
        run.style = rstyle
        set_prop_ruby(run, rubyText)
        return
    if m.groups()[0] == '':
        new_run = insert_next(p, run)
        new_run.text = m.groups()[3]
        new_run.style = rstyle
        run.text = baseText
        run.style = rstyle
        set_prop_ruby(run, rubyText)
        rubyRun_divide(p, new_run, ptn)
        return
    new_run = insert_next(p, run)
    new_run.text = baseText
    new_run.style = rstyle
    run.text = m.groups()[0]
    run.style = rstyle
    set_prop_ruby(new_run, rubyText)
    if m.groups()[3] != '':
        run = new_run
        new_run = insert_next(p, run)
        new_run.text = m.groups()[3]
        new_run.style = rstyle
        rubyRun_divide(p, new_run, ptn)

# function: divide a run ('part0 #[pr]{ part1')
def rubyRun_divide1(p, rn, ptn):
    run = p.runs[rn]
    rtext = run.text
    rstyle = run.style
    m = re.search(ptn, rtext)
    if m.groups()[0] == '' and m.groups()[1] == '':
        p._p.remove(run._r)
        return "delete"
    if m.groups()[0] == '':
        run.text = m.groups()[1]
        run.style = rstyle
        return "after"
    if m.groups()[1] == '':
        run.text = m.groups()[0]
        run.style = rstyle
        return "before"
    new_run = insert_next(p, run)
    new_run.text = m.groups()[1]
    new_run.style = rstyle
    run.text = m.groups()[0]
    run.style = rstyle
    return "both"

# function: set properties (phonetic = ruby)
def set_prop_ruby(run, rubyText):
    ruby_run = docx.oxml.shared.OxmlElement('w:r')
    ruby_run.text = rubyText
    base_run = copy.deepcopy(run._r)
    ruby = docx.oxml.shared.OxmlElement('w:ruby')
    rt = docx.oxml.shared.OxmlElement('w:rt')
    rt.append(ruby_run)
    rubyBase = docx.oxml.shared.OxmlElement('w:rubyBase')
    rubyBase.append(base_run)
    ruby.append(rt)
    ruby.append(rubyBase)
    run.text = ''
    run._r.append(ruby)

# function: set the ruby (phonetic)
def attach_ruby(document):
    ptn = re.compile(r'(.*?)#[pr]\{([^{}]+?)\:([^{}]+?)\}(.*)')
    ptn1 = re.compile(r'(.*?)#[pr]\{([^{}]*)')
    ptn2 = re.compile(r'([^{}]*?)\}(.*)')
    for p in document.paragraphs:
        if re.search(ptn, p.text) is None:
            continue
        for run in p.runs:
            rubyRun_divide(p, run, ptn)
        while True:
            m = re.search(ptn, p.text)
            if m is None:
                break
            baseText = m.groups()[1]
            rubyText = re.sub(r'[ -/:-@{-~]+', '', m.groups()[2])
            n1 = -1
            for i in range(len(p.runs)):
                if re.search(ptn1, p.runs[i].text) is not None:
                    n1 = i
                    break
            if n1 == -1:
                break
            for i in range((n1+1), len(p.runs)):
                if re.search(ptn2, p.runs[i].text):
                    n2 = i
                    break
            dtype2 = run_divide2(p, n2, ptn2)
            dtype1 = rubyRun_divide1(p, n1, ptn1)
            if dtype2 == "delete" or dtype2 == "after":
                n2 = n2 - 1
            if dtype1 == "delete":
                n2 = n2 - 1
            elif dtype1 == "before":
                n1 = n1 + 1
            elif dtype1 == "both":
                n1 = n1 + 1
                n2 = n2 + 1
            for i in range(n1+1, n2+1):
                p._p.remove(p.runs[i]._r)
            p.runs[n1].text = baseText
            set_prop_ruby(p.runs[n1], rubyText)

# remake the Word document
def remake_docx(filename, optdct):
    if not os.path.isfile(filename):
        return
    if optdct['RemakeAfter'] == False:
        return
    document = docx.Document(filename)
    section = document.sections[0]
        # arrange page margin
    section.top_margin = Mm(35.0)
    section.bottom_margin = Mm(30.0)
    section.left_margin = Mm(30.0)
    section.right_margin = Mm(30.0)
    section.header_distance = Mm(15.0)
    section.footer_distance = Mm(17.5)
    section.gutter = Mm(0.0)
        # arrange PageSize
    sz = _sizeDict[optdct['PageSize']]
    if optdct['portrait'] == False:
        w = sz[0]
        sz[0] = sz[1]
        sz[1] = w
        section.orientation = WD_ORIENT.LANDSCAPE
    section.page_width = Mm(sz[0])
    section.page_height = Mm(sz[1])
        # arrange UnderLine or EmphasisMark
    attach_ue(document)
        # arrange ruby (phonetic)
    attach_ruby(document)
        # arrange tables
    for tbl in document.tables:
        arrange_table(tbl, optdct)
    if optdct['TableCenter'] == True:
        for para in document.paragraphs:
            if para.style.name == "Table Caption":
                para.alignment = WD_ALIGN_PARAGRAPH.CENTER
        # set the standard font size
    if not optdct['FontSize'] == 12.0:
        font = document.styles['Normal'].font
        font.size = Pt(optdct['FontSize'])
        # set the text columns
    if optdct['TextColumns'] > 1:
        section = document.sections[0]
        sectPr = section._sectPr
        cols = sectPr.xpath('./w:cols')[0]
        cols.set(qn('w:num'), '{}'.format(optdct['TextColumns']))
        cols.set(qn('w:sep'), '1')  # separate line: 0 or 1
        cols.set(qn('w:equalWidth'), '1')  # equal width: 0 or 1
        # centering the paragraph with an image only
    if optdct['ImageCenter'] == True:
        inline_tag = './w:drawing/wp:inline'
        for para in document.paragraphs:
            if para.style.name == "Image Caption":
                para.alignment = WD_ALIGN_PARAGRAPH.CENTER
                continue
            if para.text == "" and len(para.runs) == 1 and \
                len(para.runs[0].element.xpath(inline_tag)) > 0:
                para.alignment = WD_ALIGN_PARAGRAPH.CENTER
        # save the document
    document.save(filename)

## main
def main():
        # コマンドライン引数のチェック
    args = sys.argv
    if len(args) < 2:
        print(
            'usage : ' + args[0] +
            ' [--ext=docx|html|pptx] infile')
        sys.exit(1)
    ext = 'docx'  # 出力ファイルの拡張子(default)
    optList = []
    fileList = []
    notList = []
    for i in range(1, len(args)):
        arg = args[i]
        if arg[0] == '-':
            if arg[0:6] == '--ext=':
                ext = arg[6:len(arg)]
            else:
                optList.append(arg)
            continue
        wlist = glob.glob(arg)
        count = 0
        for ff in wlist:
            if os.path.isfile(ff):
                fileList.append(ff)
                count = count + 1
        if count == 0:
            notList.append(arg)
    if len(notList) > 0:
        sys.stderr.write("cannot find '%s'\n" % ", ".join(notList))
    if len(fileList) < 1:
        sys.exit(1)
        # pandocのデータディレクトリ
    pandocDir = _basedir + os.sep + "pandoc"
    if not os.path.isdir(pandocDir):
        pandocDir = ''
        pandocDirOpt = ''
    else:
        pandocDirOpt = "--data-dir=" + pandocDir
    cmdInit = ['pandoc', '-s', '-f', 'markdown']
    if not pandocDirOpt == '':
        cmdInit.append(pandocDirOpt)
        # 各ファイルの処理
    file_count = 0
    for infile in fileList:
        outfile = os.path.splitext(infile)[0] + '.' + ext
        utf8file = os.path.splitext(infile)[0] + '.$utf8'
        optDict = get_options(infile)
        if optDict['PageNumber'] == True and pandocDir != '':
            optDict['pandoc'] = "--reference-doc=" + pandocDir + \
                os.sep + "reference_pn.docx " + optDict['pandoc']
        wstr = read_text(infile)
        wstr = add_heading_id(wstr)
        write_text(utf8file, wstr, 'utf-8')
        cmd = copy.copy(cmdInit)
        cmd.extend([utf8file, '-o', outfile])
        if not optDict['pandoc'] == '':
            cmd.extend(optDict['pandoc'].split())
        if len(optList) > 0:
            cmd.extend(optList)
        if os.path.isfile(outfile):
            os.unlink(outfile)
        file_count = file_count + 1
        print("#%d  %s" % (file_count, infile))
        try:
            rcode = subprocess.check_call(cmd)
        except:
            wstr = "Error on '%s'\n" % infile
            wstr = wstr + " ".join(cmd) + "\n\n"
            wstr = wstr.replace(utf8file, infile)
            sys.stderr.write(wstr)
        os.unlink(utf8file)
        if ext == 'docx':
            remake_docx(outfile, optDict)

if __name__ == '__main__':
    main()
